
//	standard includes
#include "stdafx.h"
#include	"math.h"

//	my includes
#include "covdesc.h"
#include "myarrayutils.h"


/*************************************************************************************************/
//------------------------------------------------------------------------------------------------
// PRIVATE FUNCTIONS
//------------------------------------------------------------------------------------------------
/*************************************************************************************************/




/*************************************************************************************************/
/*************************************************************************************************/
/*************************************************************************************************/






/*************************************************************************************************/
//------------------------------------------------------------------------------------------------
// PUBLIC FUNCTIONS
//------------------------------------------------------------------------------------------------
/*************************************************************************************************/
// CONSTRUCTORS

CovarianceDescriptor::CovarianceDescriptor()
{

}


CovarianceDescriptor::CovarianceDescriptor(int d)
{
	dim = d;

	numPix = 0;

	cov.set_size(dim,dim);

	invcov.set_size(dim,dim);

	invsqrtcov.set_size(dim,dim);

}


CovarianceDescriptor::CovarianceDescriptor(vnl_matrix<double> data)
{
	// data is a ( d x N ) matrix of N data points as column vectors (non-centered)

	dim = data.rows();

	numPix = data.columns();

	vnl_vector<double> rowvec(numPix);

	// centering the data matrix
	for (int i=0;i<dim;i++)
	{
		rowvec = data.get_row(i);
		data.set_row(i,rowvec - rowvec.mean());
	}
	cov = data * data.transpose()/numPix;


	eig = new vnl_symmetric_eigensystem<double>(cov);
	invcov = eig->pinverse();
	invsqrtcov = eig->inverse_square_root();

}



CovarianceDescriptor::CovarianceDescriptor(vnl_matrix<double> C, long int N)
{
	// C is the ( d x d ) covariance matrix

	dim = C.rows();

	numPix = N;

	cov = C;

	eig = new vnl_symmetric_eigensystem<double>(cov);
	invcov = eig->pinverse();
	invsqrtcov = eig->inverse_square_root();

}
/*************************************************************************************************/


// DESTRUCTOR
CovarianceDescriptor::~CovarianceDescriptor()
{

}



/*************************************************************************************************/


void CovarianceDescriptor::ComputeDescriptor(vnl_matrix<double> data)
{
	// data is a ( d x N ) matrix of N data points as column vectors (non-centered)

	dim = data.rows();

	numPix = data.columns();

	vnl_vector<double> rowvec(numPix);

	// centering the data matrix
	for (int i=0;i<dim;i++)
	{
		rowvec = data.get_row(i);
		data.set_row(i,rowvec - rowvec.mean());
	}
	cov = data * data.transpose()/numPix;

	eig = new vnl_symmetric_eigensystem<double>(cov);
	invcov = eig->pinverse();
	invsqrtcov = eig->inverse_square_root();

}


void CovarianceDescriptor::ComputeDescriptor(vnl_matrix<double> C, long int N)
{
	// C is the ( d x d ) covariance matrix

	dim = cov.rows();

	numPix = N;

	cov = C;

	eig = new vnl_symmetric_eigensystem<double>(cov);
	invcov = eig->pinverse();
	invsqrtcov = eig->inverse_square_root();

}

void CovarianceDescriptor::ComputeDescriptor(vnl_vector<double> c, int dims)
{
	// C is the ( d x d ) covariance matrix

	dim = dims;

	numPix = 0;

	vnl_sym_matrix<double> C(c.data_block(),dim);
	cov = C.as_matrix();

	eig = new vnl_symmetric_eigensystem<double>(cov);
	invcov = eig->pinverse();
	invsqrtcov = eig->inverse_square_root();

}

/*************************************************************************************************/
/*************************************************************************************************/
// Riemannian metric on tangent space of X
double CovarianceDescriptor::ComputeRiemannianMetric(vnl_matrix<double> Y, vnl_matrix<double> Z)
{
	return vnl_trace(Y * invcov * Z * invcov);
}


/*************************************************************************************************/
// Exponential map associated to Riemannian metric
vnl_matrix<double> CovarianceDescriptor::ExponentialMap(vnl_matrix<double> Y)
{
	return eig->square_root() * vnl_matrix_exp(invsqrtcov * Y * invsqrtcov) * eig->square_root();
}

/*************************************************************************************************/

// Logarithmic map associated to Riemannian metric
vnl_matrix<double> CovarianceDescriptor::LogarithmicMap(vnl_matrix<double> Y)
{
	vnl_symmetric_eigensystem<double> genEig(invsqrtcov * Y * invsqrtcov);
	genEig.D.set( cwiseLog( genEig.D.diagonal() ) );
	return (eig->square_root() * genEig.recompose() * eig->square_root());
}

/*************************************************************************************************/
// Geodesic distance squared to another covariance descriptor
double CovarianceDescriptor::GeodesicDistance(CovarianceDescriptor Y)
{
	vnl_symmetric_eigensystem<double> genEig( invsqrtcov * Y.cov * invsqrtcov );
	return pow( cwiseLog( genEig.D.diagonal() ).two_norm() , 2.0);
}

/*************************************************************************************************/


/*************************************************************************************************/



////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Computes the squared geodesic distance between all pairs of descriptors - matrices have to be symmetric and PSD
vnl_sym_matrix<double> ComputeGeodesicDistanceMatrix(CovarianceDescriptor* X, int lenX)
{
	
	vnl_sym_matrix<double> D(lenX,0.0); 

	for (int i=0;i<lenX;i++)
	{
		for (int j=i+1;j<lenX;j++)
		{
			vnl_symmetric_eigensystem<double> genEig( X[i].invsqrtcov * X[j].cov * X[i].invsqrtcov );	
			D(i,j) = pow( cwiseLog( genEig.D.diagonal() ).two_norm() , 2.0);
		}
	}
	return D;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*************************************************************************************************/


////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Computes the squared geodesic distance between all pairs of descriptors - matrices have to be symmetric and PSD
vnl_matrix<double> ComputeGeodesicDistanceMatrix(CovarianceDescriptor* X, int lenX, CovarianceDescriptor* Y, int lenY)
{
	
	vnl_matrix<double> D(lenX,lenY,0.0);


	for (int i=0;i<lenX;i++)
	{
		for (int j=0;j<lenY;j++)
		{
			vnl_symmetric_eigensystem<double> genEig( X[i].invsqrtcov * Y[j].cov * X[i].invsqrtcov );	
			D(i,j) = pow( cwiseLog( genEig.D.diagonal() ).two_norm() , 2.0);
		}
	}

	return D;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////