
//	standard includes
#include "stdafx.h"
#include	"math.h"

//	my includes
#include "imageannotator.h"
#include "myarrayutils.h"


/*************************************************************************************************/
//------------------------------------------------------------------------------------------------
// PRIVATE FUNCTIONS
//------------------------------------------------------------------------------------------------
/*************************************************************************************************/




/*************************************************************************************************/
/*************************************************************************************************/
/*************************************************************************************************/






/*************************************************************************************************/
//------------------------------------------------------------------------------------------------
// PUBLIC FUNCTIONS
//------------------------------------------------------------------------------------------------
/*************************************************************************************************/

// CONSTRUCTORS
ImageAnnotator::ImageAnnotator(string filename) : fname(filename)
{
	ifstream myfile (fname.c_str());

	if (myfile.is_open())
	{
		// number of data points
		myfile >> N;
		// cout << "N = " << N << endl;

		// covariance descriptors
		covdesc = new CovarianceDescriptor[N];

		// dimensions of covariance descriptor
		myfile >> dim;
		// cout << "dim = " << dim << endl;

		// dimensions of vectorized matrix
		int m = dim*(dim+1)/2;
		// cout << "m = " << m << endl;

		// number of classes
		myfile >> K;
		// cout << "K = " << K << endl;

		// create new KDEclass objects, one per class
		myKDEs = new KDEclass[K];

		// store the label names and other information
		for (int i=0;i<K;i++)
		{
			myfile >> myKDEs[i].N;
			myfile >> myKDEs[i].labelname;
			myKDEs[i].dim = dim;
			myKDEs[i].label = i;
			myKDEs[i].covdesc = new CovarianceDescriptor[myKDEs[i].N];
			myKDEs[i].N = 0;
		}

		COmat.set_size(K,K);
		for (int i=0;i<K;i++)
			for (int j=0;j<K;j++)
				myfile >> COmat(i,j);


		// matrix of labels
		labels = cvCreateMat(N,1,CV_8UC1);

		vnl_vector<double> Dv; Dv.set_size(m);	

		int Y;
		/// read in the data from file /////////////////
		for (int i=0;i<N;i++) 
		{
			// read in the label - labels in the file start from 1, so decrement once
			myfile >> Y; Y--; cvSet1D(labels,i,cvScalar(Y));

			// read in the vectorized matrix
			myfile >> Dv;
			// create a symmetric matrix from this vector for the covariance descriptor
			covdesc[i].ComputeDescriptor(Dv,dim);

			// put in pointers to the corresponding descriptors in the respective KDEclass objects 
			myKDEs[Y].covdesc[myKDEs[Y].N] = covdesc[i];

			// increment the count of descriptors for this class
			myKDEs[Y].N++;
		}
		////////////////////////////////////////////////

		for (int i=0;i<K;i++)
		{
			myKDEs[i].computeDistanceMatrix();
			myKDEs[i].learnKDEbandwidth();
		}
		////////////////////////////////////////////////

		myfile.close();
	}
	else
		cerr << "unable to open file" << endl;

}

/*************************************************************************************************/


/*************************************************************************************************/


// estimate class labels & probabilities
Prediction* ImageAnnotator::EstimateClassLabels(CovarianceDescriptor* test, int Ntest)
{
	vnl_matrix<double> probmat(Ntest,K);
	vnl_vector<double> probvec(Ntest);
	Prediction* Y = new Prediction[Ntest];

	// estimate probabilities of belonging to each KDE class
	for (int i=0;i<K;i++)
	{
		probvec = myKDEs[i].estimateKDEprobability(test,Ntest);
		probmat.set_column(i,probvec);
	}

	// normalization of probability distributions over classes to 1
	for (int i=0;i<Ntest;i++)
	{
		probvec = probmat.get_row(i);
		probvec /= probvec.sum();
		probmat.set_row(i,probvec);
		Y[i].Create(probvec);
	}

	return Y;
}
/*************************************************************************************************/



// refine labels 
vnl_vector<double> ImageAnnotator::RefinePrediction(Prediction* pred, ImageDescriptor img)
{
	int Ntest = img.numRegions;
	vnl_vector<double> predLabels(Ntest,0.0);


	/*** Assign most probable label as prior ***/
	for (int i=0;i<Ntest;i++)
		predLabels(i) = pred[i].MLlabel();


	/*** Let confident labels assist in their neighbors' labeling through co-occurrence ***/
	for (int i=0;i<Ntest;i++)
	{
		if ( pred[i].BvSB > 5 )
		{
			vnl_vector<double> P = img.nhoodgraph.get_row(i);
			for (int j=0;j<Ntest;j++)
			{
				if (j==i)
					continue;

				// if there is sufficient touching, and region J is not confident of its label
				if ( (P(j) > 0.1) && (pred[j].BvSB < 2) )
				{
					pred[j].probs = element_product(pred[j].probs,COmat.get_column((int) predLabels(i)));
					pred[j].SortLabels();
				}
			}
		}
	}


	/*** If there is a sufficient border overlap, and region I is significantly bigger ***/
	/*** and region J is not very confident, then region J gets the label of region I  ***/
	for (int i=0;i<Ntest;i++)
	{
		for (int j=i+1;j<Ntest;j++)
		{
			if (img.nhoodgraph(i,j) >= BORDER_THRESH)
			{
				if ( (img.numPix[i] > 5*img.numPix[j]) && (pred[j].BvSB < 5.0) )
					predLabels(j) = predLabels(i);
				else if ( (5*img.numPix[i] < img.numPix[j]) && (pred[i].BvSB < 5.0) )
					predLabels(i) = predLabels(j);
			}
		}		
	}

	/*** cars/vehicles should be small rounded objects - check eccentricity of blobs ***/
	for (int i=0;i<Ntest;i++)
	{
		int x = (int) predLabels(i);
		string tmp = myKDEs[x].labelname;		
		if ( !strcmp(tmp.c_str(),"car") || !strcmp(tmp.c_str(),"cars") || !strcmp(tmp.c_str(),"vehicle") || !strcmp(tmp.c_str(),"vehicles") )
		{
			if ( img.eccentricity[i] > ECCEN_THRESH )
			{				
				int j = 0;
				while (pred[i].label(j) == x)
					j++;
				predLabels(i) = (float) pred[i].label(j);
			}
		}
	}

	return predLabels;
}
/*************************************************************************************************/



// DESTRUCTOR
ImageAnnotator::~ImageAnnotator()
{

}

/*************************************************************************************************/